#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/time.h>
#include <linux/input.h>

extern "C" {
#include "fbutils.h"
#include "tslib.h"
};

struct	touchscreen_sample {
	int             x;
	int             y;
	unsigned int    pressure;
	int             x2;
	int             y2;
	int             nrfingers;
	int             gesture;
	struct timeval  tv;
};

struct	touchscreen_input {
	int             current_x;
	int             current_y;
	unsigned int    current_p;
	int             current_x2;
	int             current_y2;
	int             current_nrfingers;
	int             current_gesture;
};

struct button {
        int x, y, w, h;
        char *text;
        int flags;
#define BUTTON_ACTIVE 0x00000001
};

int palette [] =
{
        0x000000, 0xffe080, 0xffffff, 0xe0c0a0, 0x304050, 0x80b8c0
};
#define NR_COLORS (sizeof (palette) / sizeof (palette [0]))


/* [inactive] border fill text [active] border fill text */
int button_palette [6] =
{
        1, 4, 2,
        1, 5, 0
};

#define NR_BUTTONS 2
struct button buttons [NR_BUTTONS];

struct touchscreen_input  temp_input; //Filled until sync event received

int			iHandle;
int			iPenPositionX=0;
int			iPenPositionY=0;
bool			iPenDown=false;
int			iPen2PositionX=0;
int			iPen2PositionY=0;
bool			iPen2Down=false;
int			iNrFingers=0;
int			iGesture=0;
unsigned long long 	iEventTime;

int                     iLastPenPositionX=0;
int                     iLastPenPositionY=0;
bool                    iLastPenDown=false;
int                     iLastPen2PositionX=0;
int                     iLastPen2PositionY=0;
bool                    iLastPen2Down=false;

int	iSizeInputEvent;

static void sig(int sig)
{
        close_framebuffer();
        fflush(stderr);
        printf("signal %d caught\n", sig);
        fflush(stdout);
        exit(1);
}

static void button_draw (struct button *button)
{
        int s = (button->flags & BUTTON_ACTIVE) ? 3 : 0;
        rect (button->x, button->y, button->x + button->w - 1,
              button->y + button->h - 1, button_palette [s]);
        fillrect (button->x + 1, button->y + 1,
                  button->x + button->w - 2,
                  button->y + button->h - 2, button_palette [s + 1]);
        put_string_center (button->x + button->w / 2,
                           button->y + button->h / 2,
                           button->text, button_palette [s + 2]);
}

static int button_handle (struct button *button, int x, int y, bool pendown )
{
        int inside = (x >= button->x) && (y >= button->y) &&
                (x < button->x + button->w) &&
                (y < button->y + button->h);

        if (pendown == true) {
                if (inside) {
                        if (!(button->flags & BUTTON_ACTIVE)) {
                                button->flags |= BUTTON_ACTIVE;
                                button_draw (button);
                        }
                } else if (button->flags & BUTTON_ACTIVE) {
                        button->flags &= ~BUTTON_ACTIVE;
                        button_draw (button);
                }
        } else if (button->flags & BUTTON_ACTIVE) {
                button->flags &= ~BUTTON_ACTIVE;
                button_draw (button);
                return 1;
        }

        return 0;
}

static void refresh_screen ()
{
        int i;

        fillrect (0, 0, xres - 1, yres - 1, 0);
        put_string_center (xres/2, yres/4, (char*) "Multitouch test program", 1);
        put_string_center (xres/2, yres/4+20, (char*) "Touch screen to move crosshair", 2);

        for (i = 0; i < NR_BUTTONS; i++)
                button_draw (&buttons [i]);
}


int main(int argc, char* argv[])
{

    struct touchscreen_sample sample;
    struct input_event ev;

    bool stateChanged = false;
    bool touchScreenEventAvailable = false;
    int xtemp,ytemp;
    unsigned long long seconds;
    unsigned long long microSeconds;
    unsigned int i;
    int mode=0;

    iSizeInputEvent=sizeof(input_event);

    iHandle = open("/dev/input/event0", O_NONBLOCK);

    if (iHandle)
    {
      // make read() block until data is available
      int flags = fcntl(iHandle, F_GETFL, 0);
      if (flags >= 0)
      {
        fcntl(iHandle, F_SETFL, flags & ~O_NONBLOCK);
      }
      printf("TouchScreen: Switched On\n");
    }else
    {
      return -1; 
    }

    init_linear_filter();


    signal(SIGSEGV, sig);
    signal(SIGINT, sig);
    signal(SIGTERM, sig);

    if (open_framebuffer()) {
      close_framebuffer();
      exit(1);
    }

    for (i = 0; i < NR_COLORS; i++)
                setcolor (i, palette [i]);

    /* Initialize buttons */
    memset (&buttons, 0, sizeof (buttons));
    buttons [0].w = buttons [1].w = xres / 4;
    buttons [0].h = buttons [1].h = 20;
    buttons [0].x = xres / 4 - buttons [0].w / 2;
    buttons [1].x = (3 * xres) / 4 - buttons [0].w / 2;
    buttons [0].y = buttons [1].y = 10;
    buttons [0].text = (char*) "Drag";
    buttons [1].text = (char*) "Draw";

    refresh_screen ();

    printf("\n                      Time in Secs     PenDown     X1     Y1     Pen2Down     X2     Y2\n");

    put_cross(xres/3, yres/2, 2 | XORMODE);
    put_cross(xres*2/3, yres/2, 2 | XORMODE);

    iLastPenPositionX  = xres/3;
    iLastPenPositionY  = yres/2;
    iLastPen2PositionX = xres*2/3;
    iLastPen2PositionY  = yres/2;

	while(1){

    touchScreenEventAvailable = false; 
    stateChanged = false;

    int nr = read(iHandle, &ev, iSizeInputEvent);

    if ( nr >= iSizeInputEvent ) {
	
	switch (ev.type) {

                        case EV_SYN:
				printf("\n");
                                /* Fill out a new complete event */
                                sample.x = temp_input.current_x;
                                sample.y = temp_input.current_y;
                                sample.pressure = temp_input.current_p;
				sample.x2 = temp_input.current_x2;
				sample.y2 = temp_input.current_y2;
				sample.nrfingers = temp_input.current_nrfingers;
				sample.gesture = temp_input.current_gesture;
                                sample.tv = ev.time;
                                touchScreenEventAvailable = true;
                                break;
                        case EV_ABS:
                                switch (ev.code) {
                                case ABS_X:
                                        temp_input.current_x = ev.value;
                                        break;
                                case ABS_Y:
                                        temp_input.current_y = ev.value;
                                        break;
                                case ABS_PRESSURE:
                                        temp_input.current_p = ev.value;
                                        break;
				case ABS_FINGER2_X:
					temp_input.current_x2 = ev.value;
					break;
				case ABS_FINGER2_Y:
					temp_input.current_y2 = ev.value;
					break;
				case ABS_FINGERS:
					temp_input.current_nrfingers = ev.value;
                                        break;
                                case ABS_GESTURE:
                                        temp_input.current_gesture = ev.value;
                                        break;
                                }
                                break;
                        }
    }

    if (touchScreenEventAvailable)
    {

 	const unsigned long long timerTicksPerSecond = 1000; // The value is #define'd, but we need it to be type-safe
 	const unsigned long long microSecTomilliSecDivisor = 1000;
 
 	if ( sample.tv.tv_sec==0 )
 	{
 		gettimeofday(&sample.tv, NULL);
 	}
 
	seconds =      static_cast<unsigned long long>(sample.tv.tv_sec);
        microSeconds = timerTicksPerSecond * static_cast<unsigned long long>(sample.tv.tv_usec) / microSecTomilliSecDivisor;
 	//iEventTime = seconds + microSeconds;

        xtemp = sample.x; ytemp = sample.y;

        linear_filter(xtemp, ytemp, &sample.x, &sample.y);

        // Same for finger2

	xtemp = sample.x2; ytemp = sample.y2;

	linear_filter(xtemp, ytemp, &sample.x2, &sample.y2);

        int newPenPositionX = sample.x;
        int newPenPositionY = sample.y;
        int newPen2PositionX = sample.x2;
        int newPen2PositionY = sample.y2;
        int newNrFingers = sample.nrfingers;
 
        //Ignore finger position on finger release event, instead take the last position
        //of the finger on the screen
        if ( (iNrFingers == 2) && (newNrFingers == 1) ) {
 		newPen2PositionX = iPen2PositionX;
 		newPen2PositionY = iPen2PositionY;
 	} else if ( (iNrFingers == 1) && (newNrFingers == 0) ) {
 		newPenPositionX = iPenPositionX;
 		newPenPositionY = iPenPositionY;
 	}
 
	// The quantity of pens down changed
 	if ( iNrFingers != newNrFingers ) {
 		iPenDown = ( newNrFingers > 0);
 		iPen2Down = ( newNrFingers > 1);
 		iNrFingers = newNrFingers;
 		stateChanged=true;
 	}

       // The position of pen 1 changed
       if ((iPenPositionX != newPenPositionX) || (iPenPositionY != newPenPositionY))
       {
 		iPenPositionX = newPenPositionX;
 		iPenPositionY = newPenPositionY;
 		stateChanged = true;
       }

       // The position of pen 2 changed
       if ((iPen2PositionX != newPen2PositionX) || (iPen2PositionY != newPen2PositionY))
       {
 		iPen2PositionX = newPen2PositionX;
 		iPen2PositionY = newPen2PositionY;
 		stateChanged = true;
       }

    }
 
    if (stateChanged)
    {
 	const unsigned long long timerTicksPerSecond = 1000; // The value is #define'd, but we need it to be type-safe
 	const unsigned long long microSecTomilliSecDivisor = 1000;
 
 	if ( sample.tv.tv_sec==0 )
 	{
 		gettimeofday(&sample.tv, NULL);
 	}
 
	seconds =      static_cast<unsigned long long>(sample.tv.tv_sec);
        microSeconds = timerTicksPerSecond * static_cast<unsigned long long>(sample.tv.tv_usec) / microSecTomilliSecDivisor;
	printf(
         "After  Filtering:   %10lld.%06lld     %d        %03d     %03d      %d         %03d     %03d\n",
         seconds,
         microSeconds,
         iPenDown,
         iPenPositionX,
         iPenPositionY,
         iPen2Down,
         iPen2PositionX,
         iPen2PositionY
         );

	//Pen 1 is on the screen, update cross 1
	if (iPenDown){

        	if ((mode & 15) == 0){ //mode 0
			/* Hide the cross */
			put_cross(iLastPenPositionX, iLastPenPositionY, 2 | XORMODE);
			/* Show the cross */
			put_cross(iPenPositionX, iPenPositionY, 2 | XORMODE);
		} else if ( ((mode & 15) == 1) /*mode 1*/ && ((mode & 0x80000000) != 0))
			line (iLastPenPositionX, iLastPenPositionY, iPenPositionX, iPenPositionY, 2);

		iLastPenDown=iPenDown;
                iLastPenPositionX=iPenPositionX;
                iLastPenPositionY=iPenPositionY;

		mode |= 0x80000000;

	} else
		mode &= ~0x80000000;


	//Pen 2 is on the screen, update cross 2
	if ( 	(iPen2Down == true) 				&& \
		((iLastPen2PositionX != iPen2PositionX) 	|| \
		(iLastPen2PositionY != iPen2PositionY)) ){

        	if ((mode & 15) == 0){ //mode 0
			/* Hide the cross */
			put_cross(iLastPen2PositionX, iLastPen2PositionY, 2 | XORMODE);
			/* Show the cross */
			put_cross(iPen2PositionX, iPen2PositionY, 2 | XORMODE);
		} else if ( ((mode & 15) == 1) /*mode 1*/ && ((mode & 0x08000000) != 0))
                        line (iLastPen2PositionX, iLastPen2PositionY, iPen2PositionX, iPen2PositionY, 2);

		iLastPen2Down=iPen2Down;
                iLastPen2PositionX=iPen2PositionX;
                iLastPen2PositionY=iPen2PositionY;

		mode |= 0x08000000;
	} else
		mode &= ~0x08000000;

	for (i = 0; i < NR_BUTTONS; i++)
		if (button_handle (&buttons [i], iPenPositionX, iPenPositionY, iPenDown))
			switch (i) {
				case 0:
					mode = 0;
					refresh_screen ();
					put_cross(xres/3, yres/2, 2 | XORMODE);
    					put_cross(xres*2/3, yres/2, 2 | XORMODE);
    					iLastPenPositionX  = xres/3;
    					iLastPenPositionY  = yres/2;
    					iLastPen2PositionX = xres*2/3;
    					iLastPen2PositionY  = yres/2;
					break;
				case 1:
					mode = 1;
					refresh_screen ();
					break;
			}

	/*if (iPenDown == true) {
                        if (mode == 0x80000001){
                                line (iLastPenPositionX, iLastPenPositionY, iPenPositionX, iPenPositionY, 2);
                                line (iLastPen2PositionX, iLastPen2PositionY, iPen2PositionX, iPen2PositionY, 2);
			}
			iLastPenDown=iPenDown;
        		iLastPenPositionX=iPenPositionX;
        		iLastPenPositionY=iPenPositionY;
        		iLastPen2Down=iPen2Down;
        		iLastPen2PositionX=iPen2PositionX;
        		iLastPen2PositionY=iPen2PositionY;
                        mode |= 0x80000000;
                } else
                        mode &= ~0x80000000; */

    }

	}

    close(iHandle);
    close_framebuffer();

    return stateChanged;

  }


